import numpy as np
import pandas as pd
from fedot.api.main import Fedot
from sklearn.metrics import f1_score, roc_auc_score

from fedot_ind.api.main import FedotIndustrial
from fedot_ind.core.architecture.settings.pipeline_factory import KernelFeatureGenerator
from MKLpy.algorithms import CKA, MEMO, GRAM, RMKL, FHeuristic, EasyMKL
from MKLpy.callbacks import EarlyStopping
from MKLpy.scheduler import ReduceOnWorsening
from scipy.spatial.distance import pdist, squareform

from fedot_ind.core.architecture.pipelines.classification import ClassificationPipelines
from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader
from fedot_ind.core.ensemble.rank_ensembler import RankEnsemble


class KernelEnsembler(ClassificationPipelines):
    def __init__(self, train_data, test_data):
        super().__init__(train_data, test_data)
        self.n_classes = np.unique(self.train_target)
        self.feature_matrix_train = []
        self.feature_matrix_test = []
        if self.n_classes.shape[0] > 2:
            self.multiclass_strategy = 'ova'
            self.multiclass = True
        else:
            self.multiclass_strategy = 'ovr'
            self.multiclass = False
        self.kernel_params_list = {i.name: i.value for i in KernelFeatureGenerator}

    def __call__(self, pipeline_type: str = 'one_stage_kernel'):
        pipeline_dict = {'one_stage_kernel': self.__one_stage_kernel,
                         'two_stage_kernel': self.__two_stage_kernel
                         }
        return pipeline_dict[pipeline_type]

    def __convert_weights(self, kernel_model):
        kernels_weights_by_class = []
        if not self.multiclass:
            kernels_weights_by_class.append(kernel_model.solution.weights.cpu().detach().numpy())
        else:
            for n_class in self.n_classes:
                kernels_weights_by_class.append(kernel_model.solution[n_class].weights.cpu().detach().numpy())
        return pd.DataFrame(kernels_weights_by_class)

    def transform(self, kernel_params_dict: dict = None, feature_generator: str = None):

        if kernel_params_dict is None:
            kernel_params_dict = self.kernel_params_list
        if feature_generator is not None:
            kernel_params_dict = {feature_generator: kernel_params_dict[feature_generator]}

        for feature_generator, kernel_params in kernel_params_dict.items():
            for specified_params in kernel_params:
                feature_extractor, classificator, lambda_func_dict = self._init_pipeline_nodes(**specified_params)

                self.feature_matrix_train.append(feature_extractor.extract_features(self.train_features))
                self.feature_matrix_test.append(feature_extractor.extract_features(self.test_features))
        return

    def __one_stage_kernel(self, kernel_params_dict: dict = None, feature_generator: str = None):
        self.transform(kernel_params_dict, feature_generator)
        KLtr = [squareform(pdist(X=feature, metric='cosine')) for feature in self.feature_matrix_train]
        # mkl = CKA(multiclass_strategy=self.multiclass_strategy).fit(KLtr, self.train_target)
        mkl = FHeuristic(multiclass_strategy=self.multiclass_strategy).fit(KLtr, self.train_target)
        # mkl = EasyMKL(multiclass_strategy=self.multiclass_strategy).fit(KLtr, self.train_target)
        kernel_weight_matrix = self.__convert_weights(mkl)
        return kernel_weight_matrix

    def __two_stage_kernel(self, kernel_params_dict: dict = None, feature_generator: str = None):
        self.transform(kernel_params_dict, feature_generator)
        KLtr = [squareform(pdist(X=feature, metric='cosine')) for feature in self.feature_matrix_train]
        earlystop = EarlyStopping(
            KLtr,
            self.train_target,  # validation data, KL is a validation kernels list
            patience=5,  # max number of acceptable negative steps
            cooldown=1,  # how ofter we run a measurement, 1 means every optimization step
            metric='roc_auc',  # the metric we monitor
        )

        mkl = RMKL(multiclass_strategy='ova',
                   max_iter=1000,
                   learning_rate=.1,
                   callbacks=[earlystop],
                   scheduler=ReduceOnWorsening()).fit(KLtr, self.train_target)
        kernel_weight_matrix = self.__convert_weights(mkl)
        return kernel_weight_matrix


def init_kernel_ensemble(train_data,
                         test_data,
                         strategy: str = 'quantile',
                         kernel_list: dict = None):
    kernels = KernelEnsembler(train_data=train_data, test_data=test_data)
    set_of_fg = kernels('one_stage_kernel')(kernel_params_dict=kernel_list)

    train_feats = kernels.feature_matrix_train
    train_target = kernels.train_target

    test_feats = kernels.feature_matrix_test
    test_target = kernels.test_target

    return set_of_fg, train_feats, train_target, test_feats, test_target


if __name__ == '__main__':
    n_best = 3
    feature_dict = {}
    metric_list = []
    proba_dict = {}
    metric_dict = {}
    dataset_name = 'Lightning2'
    kernel_list = {'wavelet': [
        {'feature_generator_type': 'wavelet',
         'feature_hyperparams': {
             'wavelet': "mexh",
             'n_components': 2
         }},
        {'feature_generator_type': 'wavelet',
         'feature_hyperparams': {
             'wavelet': "morl",
             'n_components': 2
         }}],
        'quantile': [
            {'feature_generator_type': 'quantile',
             'feature_hyperparams': {
                 'window_mode': True,
                 'window_size': 25
             }
             },
            {'feature_generator_type': 'quantile',
             'feature_hyperparams': {
                 'window_mode': False,
                 'window_size': 40
             }
             }]
    }
    fg_names = []
    for key in kernel_list:
        for model_params in kernel_list[key]:
            fg_names.append(f'{key}_{model_params}')

    train_data, test_data = DataLoader(dataset_name).load_data()
    set_of_fg, train_feats, train_target, test_feats, test_target = init_kernel_ensemble(train_data,
                                                                                         test_data,
                                                                                         kernel_list=kernel_list)

    n_best_generators = set_of_fg.T.nlargest(n_best, 0).index
    for rank in range(n_best):
        fg_rank = n_best_generators[rank]
        train_best = train_feats[fg_rank]
        test_best = test_feats[fg_rank]
        feature_dict.update({fg_names[rank]: (test_best, test_best)})

    for model_name, feature in feature_dict.items():
        industrial = Fedot(
            # available_operations=['fast_ica', 'scaling','normalization',
            #                               'xgboost',
            #                               'rf',
            #                               'logit',
            #                               'mlp',
            #                               'knn',
            #                               'pca'],
            metric='roc_auc', timeout=5, problem='classification', n_jobs=6)

        model = industrial.fit(feature[0], train_target)
        labels = industrial.predict(feature[1])
        proba_dict.update({model_name: industrial.predict_proba(feature[1])})
        metric_dict.update({model_name: industrial.get_metrics(test_target, metric_names=['roc_auc', 'f1', 'acc'])})
    rank_ensembler = RankEnsemble(dataset_name=dataset_name,
                                  proba_dict=proba_dict,
                                  metric_dict=metric_dict)
    ensemble_result = rank_ensembler.ensemble()
    _ = 1




from itertools import chain
from typing import Optional

import numpy as np
import pandas as pd
from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters
from pandas import Index

from fedot_ind.core.models.base_extractor import BaseExtractor


class QuantileExtractor(BaseExtractor):
    """Class responsible for quantile feature generator experiment.

    Attributes:
        window_mode (bool): flag to use window or not
        window_size (int): size of window
        var_threshold (float): threshold for variance

    Example:
        To use this class you need to import it and call needed methods::
            from fedot.core.pipelines.pipeline_builder import PipelineBuilder
            from examples.fedot.fedot_ex import init_input_data
            from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader
            from fedot_ind.core.repository.initializer_industrial_models import IndustrialModels

            train_data, test_data = DataLoader(dataset_name='Ham').load_data()
            with IndustrialModels():
                pipeline = PipelineBuilder().add_node('quantile_extractor',
                                                      params={'window_size': 20,
                                                              'window_mode': True}).add_node(
                    'rf').build()
                input_data = init_input_data(train_data[0], train_data[1])
                pipeline.fit(input_data)
                features = pipeline.predict(input_data)
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.window_size = params.get('window_size', 0)
        self.stride = params.get('stride', 1)
        self.var_threshold = 0.1
        self.logging_params.update({'Wsize': self.window_size,
                                    'Stride': self.stride,
                                    'VarTh': self.var_threshold})

    def drop_features(self, predict: pd.DataFrame, columns: Index, n_components: int):
        """
        Method for dropping features with low variance
        """
        # Fill columns names for every extracted ts component
        predict = pd.DataFrame(predict,
                               columns=[f'{col}{str(i)}' for i in range(1, n_components + 1) for col in columns])

        if self.relevant_features is None:
            reduced_df, self.relevant_features = self.filter_by_var(predict, threshold=self.var_threshold)
            return reduced_df
        else:
            return predict[self.relevant_features]

    def filter_by_var(self, data: pd.DataFrame, threshold: float):
        cols = data.columns
        filtrat = {}

        for col in cols:
            if np.var(data[col].values) > threshold:
                filtrat.update({col: data[col].values.flatten()})

        return pd.DataFrame(filtrat), list(filtrat.keys())

    def _concatenate_global_and_local_feature(self, global_features: InputData,
                                              window_stat_features: InputData) -> InputData:
        if type(window_stat_features.features) is list:
            window_stat_features.features = np.concatenate(window_stat_features.features, axis=0)
            window_stat_features.supplementary_data['feature_name'] = list(
                chain(*window_stat_features.supplementary_data['feature_name']))

        window_stat_features.features = np.concatenate([global_features.features,
                                                        window_stat_features.features],
                                                       axis=0)
        window_stat_features.features = np.nan_to_num(window_stat_features.features)

        window_stat_features.supplementary_data['feature_name'] = list(
            chain(*[global_features.supplementary_data['feature_name'],
                    window_stat_features.supplementary_data['feature_name']]))
        return window_stat_features

    def extract_stats_features(self, ts: np.array) -> InputData:
        global_features = self.get_statistical_features(ts, add_global_features=True)
        if self.window_size != 0:
            window_stat_features = self.apply_window_for_stat_feature(ts_data=ts,
                                                                      feature_generator=self.get_statistical_features,
                                                                      window_size=self.window_size)
        else:
            window_stat_features = self.get_statistical_features(ts)

        return self._concatenate_global_and_local_feature(global_features, window_stat_features)

    def generate_features_from_ts(self,
                                  ts: np.array,
                                  window_length: int = None) -> InputData:

        ts = np.nan_to_num(ts)
        if len(ts.shape) == 2 and ts.shape[1] == 1:
            ts = ts.flatten()

        if len(ts.shape) == 1:
            aggregation_df = self.extract_stats_features(ts)
        else:
            aggregation_df = self._get_feature_matrix(self.extract_stats_features, ts)

        return aggregation_df



import logging
from typing import List, Union
from typing import Optional

import numpy as np
import pandas as pd
from fedot.api.main import Fedot
from fedot.core.data.data import array_to_input_data
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline

from fedot_ind.api.utils.checkers_collections import DataCheck
from fedot_ind.api.utils.input_data import init_input_data
from fedot_ind.api.utils.saver_collections import ResultSaver
from fedot_ind.core.metrics.evaluation import PerformanceAnalyzer


class TimeSeriesClassifier:
    """Class responsible for interaction with Fedot classifier.

    Args:
        params: which are ``generator_name``, ``generator_runner``, ``model_hyperparams``,
                ``ecm_model_flag``, ``dataset_name``, ``output_dir``.


    Attributes:
        logger (Logger): logger instance
        predictor (Fedot): Fedot model instance
        y_train (np.ndarray): target for training
        train_features (pd.DataFrame): features for training

    """

    def __init__(self, params: Optional[OperationParameters] = None):
        self.strategy = params.get('strategy')
        self.model_hyperparams = params.get('model_params')
        self.generator_runner = params.get('generator_class')
        self.dataset_name = params.get('dataset')
        self.output_folder = params.get('output_folder', None)

        self.saver = ResultSaver(dataset_name=self.dataset_name,
                                 generator_name=self.strategy,
                                 output_dir=self.output_folder)
        self.logger = logging.getLogger('TimeSeriesClassifier')
        self.datacheck = DataCheck()

        self.prediction_proba = None
        self.test_predict_hash = None
        self.prediction_label = None
        self.predictor = None
        self.y_train = None
        self.train_features = None
        self.test_features = None
        self.input_test_data = None

        self.logger.info('TimeSeriesClassifier initialised')

    def fit(self, features: Union[np.ndarray, pd.DataFrame],
            target: np.ndarray,
            **kwargs) -> object:

        baseline_type = kwargs.get('baseline_type', None)
        self.logger.info(f'Fitting model')
        self.y_train = target
        if self.generator_runner is None:
            raise AttributeError('Feature generator is not defined')

        input_data = init_input_data(features, target)
        output_data = self.generator_runner.transform(input_data)
        train_features = pd.DataFrame(output_data.predict, columns=self.generator_runner.relevant_features)

        self.train_features = self.datacheck.check_data(input_data=train_features, return_df=True)

        if baseline_type is not None:
            self.predictor = self._fit_baseline_model(self.train_features, target, baseline_type)
        else:
            self.predictor = self._fit_model(self.train_features, target)

        return self.predictor

    def _fit_model(self, features: pd.DataFrame, target: np.ndarray) -> Fedot:
        """Fit Fedot model with feature and target.

        Args:
            features: features for training
            target: target for training

        Returns:
            Fitted Fedot model

        """
        self.predictor = Fedot(**self.model_hyperparams)
        self.predictor.fit(features, target)
        self.logger.info(
            f'Solver fitted: {self.strategy}_extractor -> fedot_pipeline ({self.predictor.current_pipeline})')
        return self.predictor

    def _fit_baseline_model(self, features: pd.DataFrame, target: np.ndarray, baseline_type: str = 'rf') -> Pipeline:
        """Returns pipeline with the following structure:
            ``[initial data] -> [scaling] -> [baseline type]``

        By default, baseline type is random forest, but it can be changed to any other model from Fedot library:
        logit, knn, svc, qda, xgboost.

        Args:
            features: features for training
            target: target for training
            baseline_type: type of baseline model

        Returns:
            Fitted Fedot pipeline with baseline model

        """
        self.logger.info(f'Baseline model pipeline: scaling -> {baseline_type}. Fitting...')
        node_scaling = PrimaryNode('scaling')
        node_final = SecondaryNode(baseline_type,
                                   nodes_from=[node_scaling])
        baseline_pipeline = Pipeline(node_final)
        input_data = init_input_data(features, target)
        baseline_pipeline.fit(input_data)
        self.logger.info(f'Baseline model has been fitted')
        return baseline_pipeline

    def predict(self, features: np.ndarray, **kwargs) -> np.ndarray:
        self.prediction_label = self.__predict_abstraction(test_features=features, mode='labels', **kwargs)
        return self.prediction_label

    def predict_proba(self, features: np.ndarray, **kwargs) -> np.ndarray:
        self.prediction_proba = self.__predict_abstraction(test_features=features, mode='probs', **kwargs)
        return self.prediction_proba

    def __predict_abstraction(self,
                              test_features: Union[np.ndarray, pd.DataFrame],
                              mode: str = 'labels', **kwargs):
        self.logger.info(f'Predicting with {self.strategy} generator')

        if self.test_features is None:
            input_data = init_input_data(test_features, kwargs.get('target'))
            output_data = self.generator_runner.transform(input_data)
            test_features = pd.DataFrame(output_data.predict, columns=self.generator_runner.relevant_features)
            self.test_features = self.datacheck.check_data(input_data=test_features, return_df=True)

        if isinstance(self.predictor, Pipeline):
            self.input_test_data = init_input_data(self.test_features, kwargs.get('target'))
            prediction_label = self.predictor.predict(self.input_test_data, output_mode=mode).predict
            return prediction_label
        else:
            if mode == 'labels':
                prediction_label = self.predictor.predict(self.test_features)
            else:
                prediction_label = self.predictor.predict_proba(self.test_features)
            return prediction_label

    def get_metrics(self, target: Union[np.ndarray, pd.Series], metric_names: Union[str, List[str]]):
        analyzer = PerformanceAnalyzer()
        return analyzer.calculate_metrics(target=target,
                                          predicted_labels=self.prediction_label,
                                          predicted_probs=self.prediction_proba,
                                          target_metrics=metric_names)

    def save_prediction(self, predicted_data: np.ndarray, kind: str):
        self.saver.save(predicted_data, kind)

    def save_metrics(self, metrics: dict):
        self.saver.save(metrics, 'metrics')



from fedot_ind.core.models.quantile.stat_features import *

stat_methods = {'mean_': np.mean,
                'median_': np.median,
                'std_': np.std,
                'max_': np.max,
                'min_': np.min,
                'q5_': q5,
                'q25_': q25,
                'q75_': q75,
                'q95_': q95,
                # 'sum_': np.sum,
                # 'dif_': diff
                }

stat_methods_global = {
    'skewness_': skewness,
    'kurtosis_': kurtosis,
    'n_peaks_': n_peaks,
    'slope_': slope,
    'ben_corr_': ben_corr,
    'interquartile_range_': interquartile_range,
    'energy_': energy,
    'cross_rate_': zero_crossing_rate,
    'autocorrelation_': autocorrelation,
    # 'base_entropy_': base_entropy,
    'shannon_entropy_': shannon_entropy,
    'ptp_amplitude_': ptp_amp,
    'mean_ptp_distance_': mean_ptp_distance,
    'crest_factor_': crest_factor,
    'mean_ema_': mean_ema,
    'mean_moving_median_': mean_moving_median,
    'hjorth_mobility_': hjorth_mobility,
    'hjorth_complexity_': hjorth_complexity,
    'hurst_exponent_': hurst_exponent,
    'petrosian_fractal_dimension_': pfd,
}

# class StatFeaturesExtractor:
#     """Class for generating quantile features for a given time series.
#
#     """
#
#     @staticmethod
#     def create_baseline_features(time_series: Union[pd.DataFrame, np.ndarray]) -> pd.DataFrame:
#         """
#         Method for creating baseline quantile features for a given time series.
#
#         Args:
#             time_series: time series for which features are generated
#
#         Returns:
#             Row vector of quantile features in the form of a pandas DataFrame
#
#         """
#         names = []
#         vals = []
#         # flatten time series
#         if isinstance(time_series, (pd.DataFrame, pd.Series)):
#             time_series = time_series.values
#         time_series = time_series.flatten()
#
#         for name, method in stat_methods.items():
#             try:
#                 vals.append(method(time_series))
#                 names.append(name)
#             except ValueError:
#                 continue
#         return pd.DataFrame([vals], columns=names)



from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from sklearn.preprocessing import MinMaxScaler


class CURDecomposition:
    def __init__(self, rank):
        self.selection_rank = None
        self.rank = rank
        self.column_indices = None
        self.row_indices = None
        
    def _get_selection_rank(self, rank, matrix):
        """
        Compute the selection rank for the CUR decomposition. It must be at least 4 times the rank of the matrix but not
        greater than the number of rows or columns of the matrix.
        
        Args:
            rank: the rank of the matrix.
            matrix: the matrix to decompose.
            
        Returns:
            the selection rank
        """
        
        return min(4 * rank, min(matrix.shape))
    
    def fit_transform(self, matrix: np.ndarray) -> tuple:
        self.selection_rank = self._get_selection_rank(self.rank, matrix)

        array = np.array(matrix.copy())
        c, w, r = self.select_rows_cols(array)

        X, Sigma, y_T = np.linalg.svd(w)
        Sigma_plus = np.linalg.pinv(np.diag(Sigma))

        u = y_T.T @ Sigma_plus @ Sigma_plus @ X.T
        return c, u, r

    def reconstruct_basis(self, C, U, R, ts_length):
        # if len(U.shape) > 1:
        #     multi_reconstruction = lambda x: self.reconstruct_basis(C=C, U=U, R=x, ts_length=ts_length)
        #     TS_comps = list(map(multi_reconstruction, R))
        # else:
        rank = U.shape[1]
        TS_comps = np.zeros((ts_length, rank))
        for i in range(rank):
            X_elem = np.outer(C @ U[:, i], R[i, :])
            X_rev = X_elem[::-1]
            eigenvector = [X_rev.diagonal(j).mean() for j in range(-X_rev.shape[0] + 1, X_rev.shape[1])]
            TS_comps[:, i] = eigenvector
        return TS_comps

    def select_rows_cols(self, matrix: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        
        col_norms = np.sum(matrix ** 2, axis=0)
        row_norms = np.sum(matrix ** 2, axis=1)

        matrix_norm = np.sum(np.power(matrix, 2))

        # Compute the probabilities for selecting columns and rows
        col_probs = col_norms / matrix_norm
        row_probs = row_norms / matrix_norm

        # Select k columns and rows based on the probabilities p and q
        # selected_cols = np.random.choice(matrix.shape[1], size=self.rank, replace=False, p=col_probs)
        # selected_rows = np.random.choice(matrix.shape[0], size=self.rank, replace=False, p=row_probs)
        #
        selected_cols = np.sort(np.argsort(col_probs)[-self.selection_rank:])
        selected_rows = np.sort(np.argsort(row_probs)[-self.selection_rank:])
        # selected_cols = np.argsort(col_probs)[-self.rank:]
        # selected_rows = np.argsort(row_probs)[-self.rank:]

        row_scale_factors = 1 / np.sqrt(self.selection_rank * row_probs[selected_rows])
        col_scale_factors = 1 / np.sqrt(self.selection_rank * col_probs[selected_cols])

        C_matrix = matrix[:, selected_cols] * col_scale_factors
        R_matrix = matrix[selected_rows, :] * row_scale_factors[:, np.newaxis]
        W_matrix = matrix[selected_rows, :][:, selected_cols]
        return C_matrix, W_matrix, R_matrix

    @staticmethod
    def ts_to_matrix(time_series: np.ndarray, window: int) -> np.ndarray:
        """Make matrix from ts using window"""

        matrix = np.zeros((len(time_series) - window + 1, window))
        for i in range(len(time_series) - window + 1):
            matrix[i] = time_series[i:i + window]
        return matrix

    @staticmethod
    def matrix_to_ts(matrix: np.ndarray) -> np.ndarray:
        """Make ts from matrix"""

        ts = np.zeros(matrix.shape[0] + matrix.shape[1] - 1)
        for i in range(matrix.shape[0]):
            ts[i:i + matrix.shape[1]] += matrix[i]
        return ts

def get_random_sparse_matrix(size: tuple):
    """Generate random sparse matrix with size = size"""

    matrix = np.zeros(size)
    for i in range(size[0]):
        for j in range(size[1]):
            if np.random.rand() < 0.1:
                matrix[i, j] = np.random.rand()
    return matrix


if __name__ == '__main__':
    from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader
    from tqdm import tqdm

    arr = np.array([[1, 1, 1, 0, 0],
                    [3, 3, 3, 0, 0],
                    [4, 4, 4, 0, 0],
                    [5, 5, 5, 0, 0],
                    [0, 0, 0, 4, 4],
                    [0, 0, 0, 5, 5],
                    [0, 0, 0, 2, 2]])

    (X_train, y_train), (X_test, y_test) = DataLoader('Lightning7').load_data()

    # init_ts = train[0].iloc[0, :].values
    # scaler = MinMaxScaler()
    # scaler.fit(init_ts.reshape(-1, 1))
    # single_ts = scaler.transform(init_ts.reshape(-1, 1)).reshape(-1)

    cur = CURDecomposition(rank=20)
    # M = cur.ts_to_matrix(single_ts, 30)
    C, U, R = cur.fit_transform(X_train)
    basis = cur.reconstruct_basis(C, U, R, X_train.shape[1])

    # rec_ts = cur.matrix_to_ts(C @ U @ R)
    # err = np.linalg.norm(single_ts - rec_ts)

    # plt.plot(init_ts, label='init_ts')
    # plt.plot(scaler.inverse_transform(rec_ts.reshape(-1, 1)), label='rec_ts')
    # plt.legend()
    # plt.show()
    _ = 1

    # ranks = list(range(5, 20))
    # cur_errors = []
    # with tqdm(total=len(ranks), desc='cur') as pbar:
    #     for rank in ranks:
    #         cur = CURDecomposition(rank=rank)
    #         C, U, R = cur.fit_transform(M)
    #         cur_errors.append(np.linalg.norm(M - C @ U @ R))
    #         pbar.update(1)


    # f,a = plt.subplots(2, 1, figsize=(10, 10))
    # # a[0].plot(ranks, svd_errors, label='svd')
    # a[1].plot(ranks, cur_errors, label='cur')
    # a[0].set_title('svd')
    # a[1].set_title('cur')
    # plt.legend()
    # plt.show()
    _ = 1



from typing import Optional

from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum

from fedot_ind.core.metrics.metrics_implementation import *
from fedot_ind.core.models.base_extractor import BaseExtractor
from fedot_ind.core.models.recurrence.sequences import RecurrenceFeatureExtractor
from fedot_ind.core.operation.transformation.data.hankel import HankelMatrix
from fedot_ind.core.operation.transformation.data.kernel_matrix import TSTransformer


class RecurrenceExtractor(BaseExtractor):
    """Class responsible for wavelet feature generator experiment.

    Attributes:
        transformer: TSTransformer object.
        self.extractor: RecurrenceExtractor object.
        self.window_mode: bool, if True, then the window mode is used.
        self.min_signal_ratio: float, the minimum signal ratio.
        self.max_signal_ratio: float, the maximum signal ratio.
        self.rec_metric: str, the metric for calculating the recurrence matrix.
        self.window_size: int, the window size.

    Example:
        To use this operation you can create pipeline as follows::
            from fedot.core.pipelines.pipeline_builder import PipelineBuilder
            from examples.fedot.fedot_ex import init_input_data
            from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader
            from fedot_ind.core.repository.initializer_industrial_models import IndustrialModels

            train_data, test_data = DataLoader(dataset_name='Ham').load_data()
            with IndustrialModels():
                pipeline = PipelineBuilder().add_node('eigen_basis').add_node('recurrence_extractor').add_node(
                    'rf').build()
                input_data = init_input_data(train_data[0], train_data[1])
                pipeline.fit(input_data)
                features = pipeline.predict(input_data)
                print(features)
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.image_mode = False
        self.window_size = params.get('window_size', 0)
        self.stride = params.get('stride', 1)
        self.rec_metric = params.get('rec_metric', 'cosine')
        self.rec_metric = 'cosine' # TODO add threshold for other metrics
        self.transformer = TSTransformer
        self.extractor = RecurrenceFeatureExtractor

    def _generate_features_from_ts(self, ts: np.array):
        if self.window_size != 0:
            trajectory_transformer = HankelMatrix(time_series=ts,
                                                  window_size=self.window_size,
                                                  strides=self.stride)
            ts = trajectory_transformer.trajectory_matrix
            self.ts_length = trajectory_transformer.ts_length

        specter = self.transformer(time_series=ts,
                                   rec_metric=self.rec_metric)
        feature_df = specter.ts_to_recurrence_matrix()

        if not self.image_mode:
            feature_df = self.extractor(recurrence_matrix=feature_df).quantification_analysis()

        features = np.nan_to_num(np.array(list(feature_df.values())))
        recurrence_features = InputData(idx=np.arange(len(features)),
                                        features=features,
                                        target='no_target',
                                        task='no_task',
                                        data_type=DataTypesEnum.table,
                                        supplementary_data={'feature_name': list(feature_df.keys())})
        return recurrence_features

    def generate_recurrence_features(self, ts: np.array) -> InputData:

        if len(ts.shape) == 1:
            aggregation_df = self._generate_features_from_ts(ts)
        else:
            aggregation_df = self._get_feature_matrix(self._generate_features_from_ts, ts)

        return aggregation_df

    def generate_features_from_ts(self, ts_data: np.array,
                                  dataset_name: str = None):
        return self.generate_recurrence_features(ts=ts_data)



from typing import Union
import numpy as np
import pandas as pd
from ripser import Rips, ripser
from scipy import sparse
from fedot_ind.core.operation.transformation.data.hankel import HankelMatrix
from fedot.core.data.data import InputData, OutputData


class TopologicalTransformation:
    """Decomposes the given time series with a singular-spectrum analysis. Assumes the values of the time series are
    recorded at equal intervals.

    Args:
        time_series: Time series to be decomposed.
        max_simplex_dim: Maximum dimension of the simplices to be used in the Rips filtration.
        epsilon: Maximum distance between two points to be considered connected by an edge in the Rips filtration.
        persistence_params: ...
        window_length: Length of the window to be used in the rolling window function.

    Attributes:
        epsilon_range (np.ndarray): Range of epsilon values to be used in the Rips filtration.

    """

    def __init__(self,
                 time_series:  np.ndarray = None,
                 max_simplex_dim: int = None,
                 epsilon: int = 10,
                 persistence_params: dict = None,
                 window_length: int = None,
                 stride: int = 1):
        self.time_series = time_series
        self.stride = stride
        self.max_simplex_dim = max_simplex_dim
        self.epsilon_range = self.__create_epsilon_range(epsilon)
        self.persistence_params = persistence_params

        if self.persistence_params is None:
            self.persistence_params = {
                'coeff': 2,
                'do_cocycles': False,
                'verbose': False}

        self.__window_length = window_length

    @staticmethod
    def __create_epsilon_range(epsilon):
        return np.array([y * float(1 / epsilon) for y in range(epsilon)])

    @staticmethod
    def __compute_persistence_landscapes(ts):

        N = len(ts)
        I = np.arange(N - 1)
        J = np.arange(1, N)
        V = np.maximum(ts[0:-1], ts[1::])

        # Add vertex birth times along the diagonal of the distance matrix
        I = np.concatenate((I, np.arange(N)))
        J = np.concatenate((J, np.arange(N)))
        V = np.concatenate((V, ts))

        # Create the sparse distance matrix
        D = sparse.coo_matrix((V, (I, J)), shape=(N, N)).tocsr()
        dgm0 = ripser(D, maxdim=0, distance_matrix=True)['dgms'][0]
        dgm0 = dgm0[dgm0[:, 1] - dgm0[:, 0] > 1e-3, :]

        allgrid = np.unique(dgm0.flatten())
        allgrid = allgrid[allgrid < np.inf]

        xs = np.unique(dgm0[:, 0])
        ys = np.unique(dgm0[:, 1])
        ys = ys[ys < np.inf]

    def time_series_to_point_cloud(self,
                                   input_data: np.array = None,
                                   dimension_embed=2) -> np.array:
        """Convert a time series into a point cloud in the dimension specified by dimension_embed.

        Args:
            input_data: Time series to be converted.
            dimension_embed: dimension of Euclidean space in which to embed the time series into by taking
            windows of dimension_embed length, e.g. if the time series is ``[t_1,...,t_n]`` and dimension_embed
            is ``2``, then the point cloud would be ``[(t_0, t_1), (t_1, t_2),...,(t_(n-1), t_n)]``

        Returns:
            Collection of points embedded into Euclidean space of dimension = dimension_embed, constructed
            in the manner explained above.

        """

        if self.__window_length is None:
            self.__window_length = dimension_embed

        trajectory_transformer = HankelMatrix(time_series=input_data,
                                              window_size=self.__window_length,
                                              strides=self.stride)
        return trajectory_transformer.trajectory_matrix

    def point_cloud_to_persistent_cohomology_ripser(self,
                                                    point_cloud: np.array = None,
                                                    max_simplex_dim: int = 1):

        # ensure epsilon_range is a numpy array
        epsilon_range = self.epsilon_range

        # build filtration
        self.persistence_params['maxdim'] = max_simplex_dim
        filtration = Rips(**self.persistence_params)

        if point_cloud is None:
            point_cloud = self.time_series_to_point_cloud()

        # initialize persistence diagrams
        diagrams = filtration.fit_transform(point_cloud)
        # Instantiate persistence landscape transformer
        # plot_diagrams(diagrams)

        # normalize epsilon distance in diagrams so max is 1
        diagrams = [np.array([dg for dg in diag if np.isfinite(dg).all()]) for diag in diagrams]
        diagrams = diagrams / max(
            [np.array([dg for dg in diag if np.isfinite(dg).all()]).max() for diag in diagrams if diag.shape[0] > 0])

        ep_ran_len = len(epsilon_range)

        homology = {dimension: np.zeros(ep_ran_len).tolist() for dimension in range(max_simplex_dim + 1)}

        for dimension, diagram in enumerate(diagrams):
            if dimension <= max_simplex_dim and len(diagram) > 0:
                homology[dimension] = np.array(
                    [np.array(((epsilon_range >= point[0]) & (epsilon_range <= point[1])).astype(int))
                     for point in diagram
                     ]).sum(axis=0).tolist()

        return homology

    def time_series_to_persistent_cohomology_ripser(self,
                                                    time_series: np.array,
                                                    max_simplex_dim: int) -> dict:
        """Wrapper function that takes in a time series and outputs the persistent homology object, along with other
        auxiliary objects.

        Args:
            time_series: Time series to be converted.
            max_simplex_dim: Maximum dimension of the simplicial complex to be constructed.

        Returns:
            Persistent homology object. Dictionary with keys in ``range(max_simplex_dim)`` and, the value ``hom[i]``
            is an array of length equal to ``len(epsilon_range)`` containing the betti numbers of the ``i-th`` homology
            groups for the Rips filtration.

        """

        homology = self.point_cloud_to_persistent_cohomology_ripser(point_cloud=time_series,
                                                                    max_simplex_dim=max_simplex_dim)
        return homology

    def time_series_rolling_betti_ripser(self, ts):

        point_cloud = self.rolling_window(array=ts, window=self.__window_length)
        homology = self.time_series_to_persistent_cohomology_ripser(point_cloud,
                                                                    max_simplex_dim=self.max_simplex_dim)
        df_features = pd.DataFrame(data=homology)
        cols = ["Betti_{}".format(i) for i in range(df_features.shape[1])]
        df_features.columns = cols
        df_features['Betti_sum'] = df_features.sum(axis=1)
        return df_features

    def rolling_window(self, array, window):
        if window <= 0:
            raise ValueError("Window size must be a positive integer.")
        if window > len(array):
            raise ValueError("Window size cannot exceed the length of the array.")
        return np.array([array[i:i + window] for i in range(len(array) - window + 1)])



import os
from pathlib import Path

PROJECT_PATH = str(Path(__file__).parent.parent.parent.parent)

# Default parameters of feature generators
PATH_TO_DEFAULT_PARAMS = os.path.join(PROJECT_PATH, 'fedot_ind/core/repository/data/default_operation_params.json')

# For results collection
DS_INFO_PATH = os.path.join(PROJECT_PATH, 'core', 'architecture', 'postprocessing', 'ucr_datasets.json')


def default_path_to_save_results() -> str:
    path = PROJECT_PATH
    save_path = os.path.join(path, 'results_of_experiments')
    return save_path



from typing import Optional

from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters

from fedot_ind.core.metrics.metrics_implementation import *
from fedot_ind.core.models.base_extractor import BaseExtractor
from fedot_ind.core.operation.transformation.basis.wavelet import WaveletBasisImplementation


class SignalExtractor(BaseExtractor):
    """Class responsible for wavelet feature generator experiment.

    Attributes:
        wavelet_basis (WaveletBasisImplementation): class to transform time series
        wavelet (str): current wavelet type
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.wavelet_basis = WaveletBasisImplementation
        self.n_components = params.get('n_components')
        self.wavelet = params.get('wavelet')

    def _transform(self, input_data: InputData) -> np.array:
        wavelet_basis = self.wavelet_basis({'n_components': self.n_components,
                                            'wavelet': self.wavelet})
        transformed_features = wavelet_basis.transform(input_data)
        predict = self._clean_predict(transformed_features.predict)
        return predict



import numpy as np
from typing import List, Tuple


def weighted_inner_product(F_i: np.ndarray, F_j: np.ndarray, window_length: int, ts_length: int) -> float:
    """Calculate the weighted inner product of two vectors.

    Args:
        F_i: First vector.
        F_j: Second vector.
        window_length: Length of the window.
        ts_length: Total length of the time series.

    Returns:
        Weighted inner product.
    """
    first = list(np.arange(window_length) + 1)
    second = [window_length] * (ts_length - 2 * window_length)
    third = list(np.arange(window_length) + 1)[::-1]
    w = np.array(first + second + third)
    return float(w.dot(F_i * F_j))


def calculate_matrix_norms(TS_comps: np.ndarray, window_length: int, ts_length: int) -> np.ndarray:
    """Calculate matrix norms for the time series components.

    Args:
        TS_comps: The time series components.
        window_length: Length of the window.
        ts_length: Total length of the time series.

    Returns:
        Array of matrix norms.
    """
    r = []
    for i in range(TS_comps.shape[1]):
        r.append(weighted_inner_product(TS_comps[:, i], TS_comps[:, i], window_length, ts_length))
    F_wnorms = np.array(r)
    F_wnorms = F_wnorms ** -0.5
    return F_wnorms


def calculate_corr_matrix(ts_comps: np.ndarray,
                          f_wnorms: np.ndarray,
                          window_length: int,
                          ts_length: int) -> Tuple[np.ndarray, List[int]]:
    """Calculate the w-correlation matrix for the time series components.

    Args:
        ts_comps: The time series components.
        f_wnorms: Matrix norms of the time series components.
        window_length: Length of the window.
        ts_length: Total length of the time series.

    Returns:
        W-correlation matrix and a list of component indices.
    """
    Wcorr = np.identity(ts_comps.shape[1])
    for i in range(Wcorr.shape[0]):
        for j in range(i + 1, Wcorr.shape[0]):
            Wcorr[i, j] = abs(
                weighted_inner_product(ts_comps[:, i], ts_comps[:, j], window_length, ts_length) *
                f_wnorms[i] * f_wnorms[j])
            Wcorr[j, i] = Wcorr[i, j]
    return Wcorr, [i for i in range(Wcorr.shape[0])]


def combine_eigenvectors(ts_comps: np.ndarray, window_length: int) -> List[np.ndarray]:
    """Combine eigenvectors based on the w-correlation matrix for the time series.

    Args:
        ts_comps (np.ndarray): The time series components.
        window_length (int): Length of the window.

    Returns:
        List[np.ndarray]: List of combined eigenvectors.
    """
    ts_length = ts_comps.shape[0]
    F_wnorms = calculate_matrix_norms(ts_comps, window_length, ts_length)
    Wcorr, components = calculate_corr_matrix(ts_comps, F_wnorms, window_length, ts_length)
    combined_components = []
    current_group = []
    for i in range(len(components)):
        combined_components.append(np.array(current_group).sum(axis=0))
        current_group = [ts_comps[:, i]]
    combined_components.append(np.array(current_group).sum(axis=0))
    return combined_components



import os
from typing import Optional, Tuple

import numpy as np
import torch
from fedot.core.operations.operation_parameters import OperationParameters

from fedot_ind.core.architecture.experiment.TimeSeriesClassifier import TimeSeriesClassifier
from fedot_ind.api.utils.path_lib import default_path_to_save_results
from fedot_ind.core.models.nn.inception import InceptionTimeNetwork

TSCCLF_MODEL = {
    'inception_time': InceptionTimeNetwork
}


class TimeSeriesClassifierNN(TimeSeriesClassifier):
    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.device = torch.device('cuda' if params.get('gpu', False) else 'cpu')
        self.model = TSCCLF_MODEL[params.get('model', 'inception_time')].network_architecture
        self.num_epochs = params.get('num_epochs', 10)

    def _init_model_param(self, target: np.ndarray) -> Tuple[int, np.ndarray]:
        self.model_hyperparams['models_saving_path'] = os.path.join(default_path_to_save_results(), 'TSCNN',
                                                                    '../../models')
        self.model_hyperparams['summary_path'] = os.path.join(default_path_to_save_results(), 'TSCNN',
                                                              'runs')
        self.model_hyperparams['num_classes'] = np.unique(target).shape[0]

        if target.min() != 0:
            target = target - 1

        return self.num_epochs, target


import json
import logging
from enum import Enum
from typing import Union

from fedot_ind.api.utils.path_lib import PATH_TO_DEFAULT_PARAMS
from fedot_ind.core.architecture.settings.pipeline_factory import FeatureGenerator
from fedot_ind.core.models.base_extractor import BaseExtractor


class IndustrialConfigs(Enum):
    ts_classification = dict(task='ts_classification',
                             dataset=None,
                             strategy='fedot_preset',
                             model_params={'problem': 'classification',
                                           'seed': 42,
                                           'timeout': 15,
                                           'max_depth': 10,
                                           'max_arity': 4,
                                           'cv_folds': 2,
                                           'logging_level': 20,
                                           'n_jobs': -1})

    ts_regression = dict(task='ts_regression',
                         dataset=None,
                         strategy='quantile',
                         model_params={'problem': 'regression',
                                       'seed': 42,
                                       'timeout': 15,
                                       'max_depth': 10,
                                       'max_arity': 4,
                                       'cv_folds': 2,
                                       'logging_level': 20,
                                       'n_jobs': -1})

    anomaly_detection = dict(task='anomaly_detection',
                             dataset=None,
                             strategy='fedot_preset',
                             model_params={'problem': 'classification',
                                           'seed': 42})

    # image_classification = NotImplementedError
    # object_detection = NotImplementedError


class Configurator:
    """
    Class responsible for experiment configuration.

    """

    def __init__(self):

        self.logger = logging.getLogger(self.__class__.__name__)
        self.experiment_dict = None

    def _base_config(self, task: str = 'ts_classification') -> dict:
        return IndustrialConfigs[task].value

    def init_experiment_setup(self, **kwargs) -> dict:
        """Initializes the experiment setup.

        Args:
            kwargs: parameters of the experiment.

        Returns:
            Dictionary with the parameters of the experiment.

        """

        self.experiment_dict = self._base_config(task=kwargs['task'])
        fedot_config = {}
        industrial_config = {
            k: v if k not in ['timeout', 'n_jobs', 'logging_level', 'metric'] else fedot_config.update({k: v})
            for k, v in kwargs.items()}
        industrial_config['output_folder'] = kwargs['output_folder']
        self.experiment_dict.update(**industrial_config)
        self.experiment_dict['model_params'].update(**fedot_config)

        self.experiment_dict['generator_class'] = self._get_generator_class()

        self.__report_experiment_setup(self.experiment_dict)

        return self.experiment_dict

    def _get_generator_class(self) -> Union[BaseExtractor, None]:
        """Support method that combines the name of the generator with the parameters from the config file.

        Returns:
            Class of the feature generator.

        """
        generator = self.experiment_dict['strategy']
        if generator is None:
            return None
        elif generator == 'fedot_preset':
            return None
        else:
            if generator.startswith('ensemble'):
                dict_of_generators = {}
                generators_to_ensemble = generator.transform_for_fit(': ')[1].transform_for_fit(' ')
                for gen in generators_to_ensemble:
                    single_gen_class = self._extract_generator_class(gen)
                    dict_of_generators[gen] = single_gen_class
                ensemble_gen_class = FeatureGenerator['ensemble'].value(list_of_generators=dict_of_generators)
                self.feature_generator = 'ensemble'
                return ensemble_gen_class

            feature_gen_class = self._extract_generator_class(generator)
            return feature_gen_class

    def _extract_generator_class(self, generator):
        feature_gen_model = FeatureGenerator[generator].value

        with open(PATH_TO_DEFAULT_PARAMS, 'r') as file:
            _feature_gen_params = json.load(file)
            feature_gen_params = _feature_gen_params[f'{generator}_extractor']

        for param in feature_gen_params:
            feature_gen_params[param] = self.experiment_dict.get(param, feature_gen_params[param])

        feature_gen_class = feature_gen_model(feature_gen_params)
        return feature_gen_class

    def __report_experiment_setup(self, experiment_dict):
        """Prints the experiment setup."""

        top_info = ['task', 'dataset', 'strategy', 'branch_nodes', 'use_cache', 'n_jobs', 'timeout']
        label, data = [], []

        for obj in top_info:
            status = experiment_dict.get(obj, None)
            if status is not None:
                label.append(obj)
                data.append(status)

        self.logger.info('Experiment setup:')
        for i in range(len(label)):
            self.logger.info(f'{label[i]} - {data[i]}')



from typing import Any, List, Optional, Type, Union, Dict

import torch
import torch.nn as nn
from torch import Tensor
from torchvision.models.resnet import conv1x1, conv3x3


class BasicBlock(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        sizes: Dict[str, Tensor],
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
    ) -> None:
        super().__init__()
        norm_layer = nn.BatchNorm2d
        self.conv1 = conv3x3(sizes['conv1'][1], sizes['conv1'][0], stride=stride)
        self.bn1 = norm_layer(sizes['conv1'][0])
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(sizes['conv2'][1], sizes['conv2'][0])
        self.bn2 = norm_layer(sizes['conv2'][0])
        self.downsample = downsample
        self.stride = stride
        self.register_buffer('indices', torch.zeros(sizes['indices'], dtype=torch.int))

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out.index_add_(1, self.indices, identity)
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):

    expansion: int = 4

    def __init__(
        self,
        sizes: Dict[str, Tensor],
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
    ) -> None:
        super().__init__()
        norm_layer = nn.BatchNorm2d
        self.conv1 = conv1x1(sizes['conv1'][1], sizes['conv1'][0])
        self.bn1 = norm_layer(sizes['conv1'][0])
        self.conv2 = conv3x3(sizes['conv2'][1], sizes['conv2'][0], stride=stride)
        self.bn2 = norm_layer(sizes['conv2'][0])
        self.conv3 = conv1x1(sizes['conv3'][1], sizes['conv3'][0])
        self.bn3 = norm_layer(sizes['conv3'][0])
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.register_buffer('indices', torch.zeros(sizes['indices'], dtype=torch.int))

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out.index_add_(1, self.indices, identity)
        out = self.relu(out)

        return out


class PrunedResNet(nn.Module):
    """Pruned ResNet for soft filter pruning optimization.

    Args:
        block: ``'BasicBlock'`` or ``'Bottleneck'``.
        layers: Number of blocks on each layer.
        sizes: Sizes of layers.
        num_classes: Number of classes.
    """
    def __init__(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        layers: List[int],
        sizes: Dict,
    ) -> None:
        super().__init__()
        self.inplanes = 64
        self.conv1 = nn.Conv2d(sizes['conv1'][1], sizes['conv1'][0], kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(sizes['conv1'][0])
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(
            block=block,
            blocks=layers[0],
            sizes=sizes['layer1']
        )
        self.layer2 = self._make_layer(
            block=block,
            blocks=layers[1],
            sizes=sizes['layer2'],
            stride=2)
        self.layer3 = self._make_layer(
            block=block,
            blocks=layers[2],
            sizes=sizes['layer3'],
            stride=2)
        self.layer4 = self._make_layer(
            block=block,
            blocks=layers[3],
            sizes=sizes['layer4'],
            stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(sizes['fc'][1], sizes['fc'][0])

    def _make_layer(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        blocks: int,
        sizes: Dict,
        stride: int = 1,
    ) -> nn.Sequential:
        downsample = None
        if 'downsample' in sizes.keys():
            downsample = nn.Sequential(
                conv1x1(sizes['downsample'][1], sizes['downsample'][0], stride=stride),
                nn.BatchNorm2d(sizes['downsample'][0]),
            )
        layers = [
            block(
                sizes = sizes[0],
                stride=stride,
                downsample=downsample
            )
        ]
        for i in range(1, blocks):
            layers.append(block(sizes=sizes[i]))
        return nn.Sequential(*layers)

    def _forward_impl(self, x: Tensor) -> Tensor:
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)
        return x

    def forward(self, x: Tensor) -> Tensor:
        return self._forward_impl(x)


def pruned_resnet18(**kwargs: Any) -> PrunedResNet:
    """Pruned ResNet-18."""
    return PrunedResNet(BasicBlock, [2, 2, 2, 2], **kwargs)


def pruned_resnet34(**kwargs: Any) -> PrunedResNet:
    """Pruned ResNet-34."""
    return PrunedResNet(BasicBlock, [3, 4, 6, 3], **kwargs)


def pruned_resnet50(**kwargs: Any) -> PrunedResNet:
    """Pruned ResNet-50."""
    return PrunedResNet(Bottleneck, [3, 4, 6, 3], **kwargs)


def pruned_resnet101(**kwargs: Any) -> PrunedResNet:
    """Pruned ResNet-101."""
    return PrunedResNet(Bottleneck, [3, 4, 23, 3], **kwargs)


def pruned_resnet152(**kwargs: Any) -> PrunedResNet:
    """Pruned ResNet-152."""
    return PrunedResNet(Bottleneck, [3, 8, 36, 3], **kwargs)


PRUNED_MODELS = {
    "ResNet18": pruned_resnet18,
    "ResNet34": pruned_resnet34,
    "ResNet50": pruned_resnet50,
    "ResNet101": pruned_resnet101,
    "ResNet152": pruned_resnet152,
}
